#include "input.h"
#include "main.h"

/*
 *            Right hand coordinate system
 *                    y
 *                    ^
 *                    |
 *                    |
 *                    |
 *                    |
 *                    |
 *                    |
 *                    |
 *                    |
 *                    |
 *                    |
 *                    |__________________________\ x
 *                   /                           /
 *                  /
 *                 /
 *                /
 *               /
 *              /
 *            \//
 *            z
 */

#define XMAX 256               /* hypothetical image width */
#define YMAX 256               /* hypothetical image height */

static void pixelproc();

int Imagefile;
double **Zbuffer;
char **Image_RED;
char **Image_GREEN;
char **Image_BLUE;
char  *Output_Buffer;
int    Output_Buffer_len;

main()
{
   char     *Subroutine_Name = "main()";
   world    *World;
   polygon  *Polygons;
   vertex   *Vertices;
   int       Num_Polygons;
   int       Num_Vertices;
   view     *View;
   double   *Pointer;
   double  **View_Transformation_Matrix;
   int       Num_Eye_Points;
   int       i;
   int       j;
   char      String[80];
   window    Image = {0, 0, XMAX - 1, YMAX - 1};
   double  **Image_Transformation_Matrix;
   double  **Transformation_Matrix;
   char      Header[80];
   int       Index;


   if ((Zbuffer = (double **) malloc(XMAX * sizeof(double *))) == NULL)
   {
      fprintf(stderr, "Out of memory.\n");
      exit(1);
   }

   if ((Zbuffer[0] = (double *) malloc(XMAX * YMAX * sizeof(double))) == NULL)
   {
      fprintf(stderr, "Out of memory.\n");
      exit(1);
   }

   if ((Image_RED = (char **) malloc(XMAX * sizeof(char *))) == NULL)
   {
      fprintf(stderr, "Out of memory.\n");
      exit(1);
   }

   if ((Image_RED[0] = (char *) malloc(XMAX * YMAX * sizeof(char))) == NULL)
   {
      fprintf(stderr, "Out of memory.\n");
      exit(1);
   }

   if ((Image_GREEN = (char **) malloc(XMAX * sizeof(char *))) == NULL)
   {
      fprintf(stderr, "Out of memory.\n");
      exit(1);
   }

   if ((Image_GREEN[0] = (char *) malloc(XMAX * YMAX * sizeof(char))) == NULL)
   {
      fprintf(stderr, "Out of memory.\n");
      exit(1);
   }

   if ((Image_BLUE = (char **) malloc(XMAX * sizeof(char *))) == NULL)
   {
      fprintf(stderr, "Out of memory.\n");
      exit(1);
   }

   if ((Image_BLUE[0] = (char *) malloc(XMAX * YMAX * sizeof(char))) == NULL)
   {
      fprintf(stderr, "Out of memory.\n");
      exit(1);
   }

   Output_Buffer_len = 3 * YMAX * sizeof(char);
   if ((Output_Buffer = (char *) malloc(Output_Buffer_len)) == NULL)
   {
      fprintf(stderr, "Out of memory.\n");
      exit(1);
   }

   for (i = 1; i < XMAX; i++)
   {
      Zbuffer[i] = &(Zbuffer[i - 1][YMAX]);
      Image_RED[i] = &(Image_RED[i - 1][YMAX]);
      Image_GREEN[i] = &(Image_GREEN[i - 1][YMAX]);
      Image_BLUE[i] = &(Image_BLUE[i - 1][YMAX]);
      for (j = 0; j < YMAX; j++)
      {
         Zbuffer[i][j] = HUGE_VAL;
         Image_RED[i][j] = 0;
         Image_GREEN[i][j] = 0;
         Image_BLUE[i][j] = 0;
      }
   }

   World = Read_World_Data("sphere3.dat");   

   Polygons = World -> Polygons;
   Vertices = World -> Vertices;

   Num_Polygons = World -> Num_Polygons;
   Num_Vertices = World -> Max_Vertices_per_Polygon;

#if 0
   printf("Num_Polygons = %d, Max_Num_Vertices = %d\n",
          Num_Polygons, World -> Max_Vertices_per_Polygon);
   for (i = 0; i < Num_Polygons; i++)
   {
      fprintf(String, "Polygon: %d", i);
      polygon_print(String, &(Polygons[i]));
   }
#endif

   View = Read_View_Data("view.dat");   

   View_Transformation_Matrix = Get_Transformation_Matrix();

   printf("View -> Distance_to_Screen = %lf\n", View -> Distance_to_Screen);
   printf("View -> Near_Plane_Distance = %lf\n", View -> Near_Plane_Distance);
   printf("View -> Convergence_Distance = %lf\n", View -> Convergence_Distance);
   printf("View -> Screen LL (%d, %d) to UR (%d, %d)\n",
          View -> Screen.LL_x, View -> Screen.LL_y,
          View -> Screen.UR_x, View -> Screen.UR_y);
   printf("View -> Num_Eye_Points = %d\n", View -> Num_Eye_Points);

#if 0
   Num_Eye_Points = View -> Num_Eye_Points;

   for (i = 0; i < Num_Eye_Points; i++)
   {
      printf("Eye # %d\n", i);
      printf("View_direction = (%lf, %lf, %lf)\n",
             View -> Eye[i].View_direction[I_HAT],
             View -> Eye[i].View_direction[J_HAT],
             View -> Eye[i].View_direction[K_HAT]);
      printf("Up = (%lf, %lf, %lf)\n",
             View -> Eye[i].Up[I_HAT],
             View -> Eye[i].Up[J_HAT],
             View -> Eye[i].Up[K_HAT]);
      printf("Location = (%lf, %lf, %lf)\n",
             View -> Eye[i].Location[I_HAT],
             View -> Eye[i].Location[J_HAT],
             View -> Eye[i].Location[K_HAT]);
   }
#endif

   View -> Current_Eye_Point = 0;
   Set_View_Transformation_Matrix(View, View_Transformation_Matrix);

   Image_Transformation_Matrix = Get_Transformation_Matrix();

   Set_Image_Transformation_Matrix(&Image, View, Image_Transformation_Matrix);

#if 0
   Matrix_4x4_Multiplication(View_Transformation_Matrix,
                             Image_Transformation_Matrix,
                             View_Transformation_Matrix);
#endif

   Transform_World(Image_Transformation_Matrix, World);

   for (i = 0; i < Num_Polygons; i++)
   {
      /* scan convert! */
      polygon_scan_convert(&(Polygons[i]), &Image, pixelproc);
   }

   if ((Imagefile = open("test.ppm", O_CREAT | O_WRONLY | O_TRUNC,
                         S_IREAD | S_IWRITE)) == -1)
   {
      fprintf(stderr, "Cannot open output file.\n");
      exit(1);
   }

   sprintf(Header, "P6\n%d %d\n255\n\0", XMAX, YMAX);
   write(Imagefile, Header, strlen(Header));

   for (i = 0; i < XMAX; i++)
   {
      for (j = 0, Index = 0; j < YMAX; j++, Index += 3)
      {
         Output_Buffer[Index] = Image_RED[j][i];
         Output_Buffer[Index + 1] = Image_GREEN[j][i];
         Output_Buffer[Index + 2] = Image_BLUE[j][i];
      }
      write(Imagefile, Output_Buffer, Output_Buffer_len);
   }

   close(Imagefile);
}

static void pixelproc(int x, int y, vertex *point)
{
#if 0
   printf("pixel (%d,%d) screenz=%g rgb=(%g,%g,%g)\n",
      x, y, point->sz, point->r, point->g, point->b);
#endif
   /*
    * in real graphics program you could read and write pixels, e.g.:
    *
    * if (point->sz < zbuffer_read(x, y)) { image_write_rgb(x, y, point->r,
    * point->g, point->b); zbuffer_write(x, y, point->sz); }
    */

    if (point -> sz < Zbuffer[x][y])
    {
       Zbuffer[x][y] = point -> sz;
       Image_RED[x][y] = (int) floor((point->r * 255.0) + 0.5);
       Image_GREEN[x][y] = (int) floor((point->g * 255.0) + 0.5);
       Image_BLUE[x][y] = (int) floor((point->b * 255.0) + 0.5);
    }
}

